'use strict';

module.exports = makeEventsDispatcher;

/**
 * Returns an event dispatcher.
 * @param configRepository An object that provides a `get` method to get the 'defaultHandler'`
 * @param eventsRepository An object that provides a `handlers` method to retrieve all handlrs registered for an event
 * @returns {{queuedHandlers: {}, dispatch: (function(*=, *=)), queueHandlers: (function(*=, *=)), dequeueHandlers: (function(*))}}
 */
function makeEventsDispatcher(configRepository, eventsRepository) {
    return {
        queuedHandlers: {},

        dispatch: function dispatch(eventName, payload) {
            this.queueHandlers(eventName, payload);
        },
        queueHandlers: function queueHandlers(eventName, eventData) {
            var _this = this;

            var handlers = eventsRepository.handlers(eventName);

            if (handlers.length <= 0) {
                if (!configRepository.get('defaultHandler')) return;
                handlers = [configRepository.get('defaultHandler')];
            }

            this.queuedHandlers[eventName] = [];

            var _loop = function _loop(handlerMethod) {
                var handlerRef = setTimeout(function () {
                    var continuePropagation = handlerMethod(eventData);
                    // stop the next handlers if this one says so
                    if (continuePropagation === false) {
                        _this.dequeueHandlers(eventName);
                    }
                }, 0);
                _this.queuedHandlers[eventName].push(handlerRef);
            };

            var _iteratorNormalCompletion = true;
            var _didIteratorError = false;
            var _iteratorError = undefined;

            try {
                for (var _iterator = handlers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                    var handlerMethod = _step.value;

                    _loop(handlerMethod);
                }
            } catch (err) {
                _didIteratorError = true;
                _iteratorError = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion && _iterator.return) {
                        _iterator.return();
                    }
                } finally {
                    if (_didIteratorError) {
                        throw _iteratorError;
                    }
                }
            }
        },
        dequeueHandlers: function dequeueHandlers(eventName) {
            var _iteratorNormalCompletion2 = true;
            var _didIteratorError2 = false;
            var _iteratorError2 = undefined;

            try {
                for (var _iterator2 = this.queuedHandlers[eventName][Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                    var handlerRef = _step2.value;

                    clearTimeout(handlerRef);
                }
            } catch (err) {
                _didIteratorError2 = true;
                _iteratorError2 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion2 && _iterator2.return) {
                        _iterator2.return();
                    }
                } finally {
                    if (_didIteratorError2) {
                        throw _iteratorError2;
                    }
                }
            }
        }
    };
}